import os
import re
import shutil
import time

import github
import modal
import requests
from github import Github
from github.Repository import Repository
from jwt import encode
from loguru import logger
from tqdm import tqdm

from sweepai.core.entities import Snippet
from sweepai.utils.config.client import SweepConfig
from sweepai.utils.config.server import DB_MODAL_INST_NAME, GITHUB_APP_ID, GITHUB_APP_PEM
from sweepai.utils.ctags_chunker import get_ctags_for_file
from sweepai.utils.event_logger import posthog


def make_valid_string(string: str):
    pattern = r"[^\w./-]+"
    return re.sub(pattern, "_", string)


def get_jwt():
    signing_key = GITHUB_APP_PEM
    app_id = GITHUB_APP_ID
    payload = {"iat": int(time.time()), "exp": int(time.time()) + 600, "iss": app_id}

    return encode(payload, signing_key, algorithm="RS256")


def get_token(installation_id: int):
    jwt = get_jwt()
    headers = {
        "Accept": "application/vnd.github+json",
        "Authorization": "Bearer " + jwt,
        "X-GitHub-Api-Version": "2022-11-28",
    }
    response = requests.post(
        f"https://api.github.com/app/installations/{int(installation_id)}/access_tokens",
        headers=headers,
    )
    return response.json()["token"]


def get_github_client(installation_id: int):
    token = get_token(installation_id)
    return Github(token)


def get_installation_id(username: str):
    jwt = get_jwt()
    response = requests.get(
        f"https://api.github.com/users/{username}/installation",
        headers={
            "Accept": "application/vnd.github+json",
            "Authorization": "Bearer " + jwt,
            "X-GitHub-Api-Version": "2022-11-28",
        },
    )
    obj = response.json()
    try:
        return obj["id"]
    except:
        raise Exception("Could not get installation id, probably not installed")

def list_directory_tree(
    root_directory,
    included_directories=None,
    excluded_directories=None,
    included_files=None,
):
    """Display the directory tree.

    Arguments:
    root_directory -- String path of the root directory to display.
    included_directories -- List of directory paths (relative to the root) to include in the tree. Default to None.
    excluded_directories -- List of directory names to exclude from the tree. Default to None.
    """

    # Default values if parameters are not provided
    if included_directories is None:
        included_directories = []
    if excluded_directories is None:
        excluded_directories = [".git"]

    def list_directory_contents(
        current_directory,
        indentation=""
    ):
        """Recursively list the contents of directories."""

        file_and_folder_names = os.listdir(current_directory)
        file_and_folder_names.sort()

        directory_tree_string = ""

        for name in file_and_folder_names:
            relative_path = os.path.join(current_directory, name)[len(root_directory) + 1:]
            if name in excluded_directories:
                continue
            complete_path = os.path.join(current_directory, name)

            if os.path.isdir(complete_path):
                if relative_path in included_directories:
                    directory_tree_string += f"{indentation}{relative_path}/\n"
                    directory_tree_string += list_directory_contents(
                        complete_path, indentation + "  "
                    )
                else:
                    directory_tree_string += f"{indentation}{name}/...\n"
            else:
                directory_tree_string += f"{indentation}{name}\n"
                if os.path.isfile(complete_path) and relative_path in included_files:
                    ctags = get_ctags_for_file(complete_path)
                    ctags = "\n".join([indentation + line for line in ctags.splitlines()])
                    if ctags.strip():
                        directory_tree_string += f"{ctags}\n"
        return directory_tree_string

    directory_tree = list_directory_contents(root_directory)
    return directory_tree

def get_file_list(root_directory: str) -> str:
    files = []

    def dfs_helper(directory):
        nonlocal files
        for item in os.listdir(directory):
            if item == ".git":
                continue
            item_path = os.path.join(directory, item)
            if os.path.isfile(item_path):
                files.append(item_path)  # Add the file to the list
            elif os.path.isdir(item_path):
                dfs_helper(item_path)  # Recursive call to explore subdirectory

    dfs_helper(root_directory)
    files = [file[len(root_directory) + 1:] for file in files]
    return files

def get_tree_and_file_list(
        repo: Repository,
        installation_id: int,
        snippet_paths: list[str]
) -> str:
    from git import Repo
    token = get_token(installation_id)
    shutil.rmtree("repo", ignore_errors=True)
    repo_url = f"https://x-access-token:{token}@github.com/{repo.full_name}.git"
    git_repo = Repo.clone_from(repo_url, "repo")
    git_repo.git.checkout(SweepConfig.get_branch(repo))

    prefixes = []
    for snippet_path in snippet_paths:
        file_list = ""
        for directory in snippet_path.split("/")[:-1]:
            file_list += directory + "/"
            prefixes.append(file_list.rstrip("/"))
        file_list += snippet_path.split("/")[-1]
        prefixes.append(snippet_path)

    tree = list_directory_tree(
        "repo",
        included_directories=prefixes,
        included_files=snippet_paths,
    )
    file_list = get_file_list("repo")
    shutil.rmtree("repo")
    return tree, file_list


def get_file_contents(repo: Repository, file_path, ref=None):
    if ref is None:
        ref = repo.default_branch
    file = repo.get_contents(file_path, ref=ref)
    contents = file.decoded_content.decode("utf-8", errors='replace')
    return contents


def search_snippets(
        repo: Repository,
        query: str,
        installation_id: int,
        num_files: int = 5,
        include_tree: bool = True,
        branch: str = None,
        sweep_config: SweepConfig = SweepConfig(),
) -> tuple[list[Snippet], str]:
    # Initialize the relevant directories string
    get_relevant_snippets = modal.Function.lookup(DB_MODAL_INST_NAME, "get_relevant_snippets")
    snippets: list[Snippet] = get_relevant_snippets.call(
        repo.full_name, query, num_files, installation_id=installation_id
    )
    logger.info(f"Snippets: {snippets}")
    # TODO: We should prioritize the mentioned files
    for snippet in snippets:
        try:
            file_contents = get_file_contents(repo, snippet.file_path, ref=branch)
            if len(file_contents) > sweep_config.max_file_limit:  # more than ~10000 tokens
                logger.warning(f"Skipping {snippet.file_path}, too many tokens")
                continue
        except github.UnknownObjectException as e:
            logger.warning(f"Error: {e}")
            logger.warning(f"Skipping {snippet.file_path}")
        else:
            snippet.content = file_contents
    tree, file_list = get_tree_and_file_list(
        repo,
        installation_id,
        snippet_paths=[snippet.file_path for snippet in snippets]
    )
    for file_path in tqdm(file_list):
        if file_path in query:
            try:
                file_contents = get_file_contents(repo, file_path, ref=branch)
                if len(file_contents) > sweep_config.max_file_limit:  # more than 10000 tokens
                    logger.warning(f"Skipping {file_path}, too many tokens")
                    continue
            except github.UnknownObjectException as e:
                logger.warning(f"Error: {e}")
                logger.warning(f"Skipping {file_path}")
            else:
                snippets = [
                               Snippet(
                                   content=file_contents,
                                   start=0,
                                   end=file_contents.count("\n") + 1,
                                   file_path=file_path,
                               )
                           ] + snippets
    snippets = [snippet.expand() for snippet in snippets]
    if include_tree:
        return snippets, tree
    else:
        return snippets


def index_full_repository(
        repo_name: str,
        installation_id: int = None,
        sweep_config: SweepConfig = SweepConfig(),
):
    update_index = modal.Function.lookup(DB_MODAL_INST_NAME, "update_index")
    num_indexed_docs = update_index.spawn(
        repo_name=repo_name,
        installation_id=installation_id,
        sweep_config=sweep_config,
    )
    try:
        repo = get_github_client(installation_id).get_repo(repo_name)
        labels = repo.get_labels()
        label_names = [label.name for label in labels]

        if "sweep" not in label_names:
            repo.create_label(
                name="sweep",
                color="5320E7",
                description="Assigns Sweep to an issue or pull request.",
            )
    except Exception as e:
        posthog("index_full_repository", "failed", {"error": str(e)})
        logger.warning(
            "Adding label failed, probably because label already."
        )  # warn that the repo may already be indexed
    return num_indexed_docs



import modal
import subprocess
import os

from sweepai.utils.ctags import Ctags

MAX_NUM_TAGS = 5

def unified_ctags_sorter(tags):
    # Define a unified priority map for both Python and JavaScript
    priority_map = {
        'module': 1,  # Python
        'namespace': 2,  # JavaScript/TypeScript
        'class': 3,  # Python, JavaScript, TypeScript
        'interface': 4,  # TypeScript
        'function': 5,  # Python, JavaScript, TypeScript
        'method': 6,  # Python, JavaScript, TypeScript
        'variable': 7,  # Python, JavaScript, TypeScript
        'constant': 8,  # JavaScript, TypeScript
        'member': 9,  # Python, JavaScript, TypeScript
        'enum': 10,  # TypeScript
        'property': 11  # JavaScript, TypeScript
    }

    # Sort the tags by priority, fallback to their original order for equal priorities
    sorted_tags = sorted(tags, key=lambda tag: (priority_map.get(tag[0], 5), tags.index(tag)))

    # Return the first MAX_NUM_TAGS tags
    return sorted_tags[:MAX_NUM_TAGS]

def should_add_tag(tag):
    if tag['kind'] == 'variable':
        return False
    if 'scope' not in tag and 'signature' in tag and len(tag['signature']) < 10:
        return False
    return True

def get_ctags_for_file(file_path):
    repo = Ctags(file_path)
    tags = repo.run_ctags()
    tag_structure = []
    for tag in tags:
        kind = tag['kind']
        name = tag['name']
        signature = None
        if 'signature' in tag:
            signature = tag['signature']
        if should_add_tag(tag):
            tag_structure.append((kind, name, signature))
    # Organize the tags by file and kind
    tag_structure = unified_ctags_sorter(tag_structure)

    # Generate the string
    output = ""
    for (kind, name, signature) in tag_structure:
        sig = ' ' + signature if signature else ''
        output += f"  {kind} {name}{sig}\n"
    return output


import difflib
import re

def diff_contains_dups_or_removals(diff, new_code):
    # The regex pattern for lines removed or added in the actual code
    removed_line_pattern = r"^-.*"
    added_line_pattern = r"^\+.*"

    lines_removed = False
    duplicate_lines_added = False

    # Split the diff and new_code into separate lines
    diff_lines = diff.split('\n')[3:]  # Start from the third line
    new_code_lines = [line.strip() for line in new_code.split('\n')]

    # Check if there are removed lines
    for line in diff_lines:
        if re.match(removed_line_pattern, line):
            lines_removed = True

    # Check if there are duplicate lines added
    added_lines = [line[1:].strip() for line in diff_lines if re.match(added_line_pattern, line)]
    for line in added_lines:
        if new_code_lines.count(line) > 1:
            duplicate_lines_added = True
            break

    return lines_removed or duplicate_lines_added

def generate_diff(old_code, new_code):
    old_code = old_code.strip()
    new_code = new_code.strip()
    
    diff = difflib.unified_diff(
        old_code.splitlines(keepends=True),
        new_code.splitlines(keepends=True)
    )

    diff_text = ''.join(diff)
    
    return diff_text

def revert_whitespace_changes(original_file_str, modified_file_str):
    original_lines = original_file_str.splitlines()
    modified_lines = modified_file_str.splitlines()

    diff = difflib.SequenceMatcher(None, original_lines, modified_lines)

    final_lines = []
    for opcode in diff.get_opcodes():
        if opcode[0] == "equal" or opcode[0] == "replace":
            # If the lines are equal or replace (means the change is not whitespace only)
            # use original lines.
            final_lines.extend(original_lines[opcode[1]:opcode[2]])
        elif opcode[0] == "insert":
            # If the lines are inserted in the modified file, check if it's just whitespace changes
            # If it's just whitespace changes, ignore them.
            for line in modified_lines[opcode[3]:opcode[4]]:
                if line.strip() != "":
                    final_lines.append(line)

    return '\n'.join(final_lines)


def format_contents(file_contents, is_markdown=False):
    """
    Add arbitrary postprocessing here, this affects files and diffs
    """
    lines = file_contents.split('\n')

    if is_markdown:
        return '\n'.join(lines)

    # Handle small files
    if len(lines) <= 5:
        start_idx = 0
        end_idx = len(lines)
        for idx, line in enumerate(lines):
            if start_idx == 0 and line.strip().startswith('```'):
                start_idx = idx + 1
            if start_idx != 0 and line.strip().endswith('```'):
                end_idx = idx
        lines = lines[start_idx:end_idx]
        return '\n'.join(lines)

    first_three_lines = lines[:3]
    last_three_lines = lines[-3:]
    first_line_idx = 0
    last_line_idx = 3
    for idx, line in enumerate(first_three_lines):
        line = line.strip()
        if line.startswith('```'):
            first_line_idx = max(first_line_idx, idx + 1)
        if 'user_code>' in line:
            first_line_idx = max(first_line_idx, idx + 1)
    for idx, line in enumerate(last_three_lines): # Check in reverse
        line = line.strip()
        if line.endswith('```'):
            last_line_idx = min(idx, last_line_idx)
        if 'user_code>' in line:
            last_line_idx = min(idx, last_line_idx)
    first_three_lines = first_three_lines[first_line_idx:]
    last_three_lines = last_three_lines[:last_line_idx]

    lines = first_three_lines + lines[3:-3] + last_three_lines
    return '\n'.join(lines)


def generate_new_file(modify_file_response: str, old_file_content: str, chunk_offset: int=0) -> str:
    result_file = ""
    old_file_lines = old_file_content.splitlines()

    # Extract content between <new_file> tags
    new_file = re.search(r".*?<new_file>\n?(.*)\n<\/new_file>", modify_file_response, re.DOTALL).group(1)
    if "<copy_lines" not in new_file:
        return new_file

    # v5
    result = []
    lines = new_file.split('\n')
    for line_number, line in enumerate(lines):
        # Todo: make it support 1 number only
        matches = re.finditer(r"<copy_lines\s(\d+-\d+)/?>", line)
        copied_lines = False
        for match in matches:
            copied_lines = True
            start, end = match.group(1).split('-')
            start, end = int(start) - 1, int(end) - 1
            if chunk_offset != 0: # Correct for the line numbers being much higher
                start -= chunk_offset
                end -= chunk_offset
            start = max(0, start)
            end = min(len(old_file_lines) - 1, end)

            replacements = old_file_lines[start:end + 1]
            replacements_str = '\n'.join(replacements)
            line = line.replace(match.group(0), replacements_str)

        # check if line was incorrectly duplicated
        append = True
        if not copied_lines:  # if bot generated, and line before is not bot generated
            if len(result) > 0:
                # Get last line in results
                last_group = result[-1]
                # last_line = last_group
                if '\n' in last_group:
                    last_line = last_group[last_group.rindex('\n') + 1:]  # if its multiple lines
                    # if last line is same is current line
                    if last_line == line:
                        append = False

        if append:
            result.append(line)
    result = '\n'.join(result)

    return result


def sliding_window_replacement(original, search, replace):
    index = -1
    max_similarity = 0
    # sliding window comparison from original to search
    # Todo: use rapidfuzz to compute fuzzy similarity over code
    for i in range(len(original) - len(search) + 1):
        count = 0
        for j in range(len(search)):
            if search[j].strip() == original[i + j].strip():
                count += 1
        if count > max_similarity:
            index = i
            max_similarity = count

    # No changes could be found. Return original code.
    if index == -1:
        return original

    snippet = original[index:index + len(search)]

    # Fix whitespace
    spaces = ''
    if len(search[0]) - len(search[0].lstrip()) == 0:
        spaces = ' ' * (len(snippet[0]) - len(snippet[0].lstrip()))
    # Todo: What if whitespace in search is incorrect

    modified = [spaces + line for line in replace]

    # replaced original with modified
    original = original[:index] + modified + original[index + len(search):]
    return original

def generate_new_file_from_patch(modify_file_response: str, old_file_content: str, chunk_offset: int=0) -> str:
    old_file_lines = old_file_content.splitlines()

    # Extract content between <new_file> tags
    matches = re.findall(r'<<<<.*?\n(.*?)\n?====\n(.*?)\n?>>>>', modify_file_response, re.DOTALL)

    for search, replace in matches:
        old_file_lines = sliding_window_replacement(old_file_lines, search.splitlines(), replace.splitlines())

    result = '\n'.join(old_file_lines)
    return result


def join_contents_k(first, second, k):
    """
    Join contents together removing k duplicate lines
    """
    first_lines = first.splitlines()
    second_lines = second.splitlines()
    for i in range(k, 0, -1):
        if len(first_lines) < k or len(second_lines) < k:
            continue
        if first_lines[-i:] == second_lines[:i]:
            return "\n".join(first_lines) + "\n" + "\n".join(second_lines[i:])
    return "\n".join(first_lines) + "\n" + "\n".join(second_lines)


def is_markdown(filename):
    return filename.endswith(".md") or filename.endswith(".rst") or filename.endswith(".txt")



from dataclasses import dataclass
import json
import os
import subprocess
from collections import defaultdict


@dataclass
class Ctags:
    file: str
    ctags_cmd = [
        "ctags",
        "--fields=+S",
        "--extras=-F",
        "--output-format=json",
        "--output-encoding=utf-8",
    ]
    def run_ctags(self):
        cmd = self.ctags_cmd + [
            "--input-encoding=utf-8",
            self.file
        ]
        output = subprocess.check_output(cmd, stderr=subprocess.PIPE).decode("utf-8")
        output_lines = output.splitlines()
        data = []
        for line in output_lines:
            try:
                tag = json.loads(line)
                if tag["_type"] == "tag":
                    data.append(tag)
            except json.decoder.JSONDecodeError as err:
                pass
        return data



import traceback

import openai
from loguru import logger

from sweepai.core.entities import NoFilesException, Snippet
from sweepai.core.sweep_bot import SweepBot
from sweepai.handlers.on_review import get_pr_diffs
from sweepai.utils.chat_logger import ChatLogger
from sweepai.utils.config.server import PREFIX, OPENAI_API_KEY, GITHUB_BOT_TOKEN
from sweepai.utils.event_logger import posthog
from sweepai.utils.github_utils import (
    get_github_client,
    search_snippets,
)
from sweepai.utils.prompt_constructor import HumanMessageCommentPrompt

github_access_token = GITHUB_BOT_TOKEN
openai.api_key = OPENAI_API_KEY

num_of_snippets_to_query = 30
total_number_of_snippet_tokens = 15_000
num_full_files = 2
num_extended_snippets = 2


def post_process_snippets(snippets: list[Snippet], max_num_of_snippets: int = 3):
    for snippet in snippets[:num_full_files]:
        snippet = snippet.expand()

    # snippet fusing
    i = 0
    while i < len(snippets):
        j = i + 1
        while j < len(snippets):
            if snippets[i] ^ snippets[j]:  # this checks for overlap
                snippets[i] = snippets[i] | snippets[j]  # merging
                snippets.pop(j)
            else:
                j += 1
        i += 1

    # truncating snippets based on character length
    result_snippets = []
    total_length = 0
    for snippet in snippets:
        total_length += len(snippet.get_snippet())
        if total_length > total_number_of_snippet_tokens * 5:
            break
        result_snippets.append(snippet)
    return result_snippets[:max_num_of_snippets]


def on_comment(
        repo_full_name: str,
        repo_description: str,
        comment: str,
        pr_path: str | None,
        pr_line_position: int | None,
        username: str,
        installation_id: int,
        pr_number: int = None,
        comment_id: int | None = None,
):
    # Check if the comment is "REVERT"
    if comment.strip().upper() == "REVERT":
        rollback_file(repo_full_name, pr_path, installation_id, pr_number)
        return {"success": True, "message": "File has been reverted to the previous commit."}

    # Flow:
    # 1. Get relevant files
    # 2: Get human message
    # 3. Get files to change
    # 4. Get file changes
    # 5. Create PR
    logger.info(
        f"Calling on_comment() with the following arguments: {comment}, {repo_full_name}, {repo_description}, {pr_path}")
    organization, repo_name = repo_full_name.split("/")
    metadata = {
        "repo_full_name": repo_full_name,
        "repo_name": repo_name,
        "organization": organization,
        "repo_description": repo_description,
        "installation_id": installation_id,
        "username": username,
        "function": "on_comment",
        "mode": PREFIX,
    }

    posthog.capture(username, "started", properties=metadata)
    logger.info(f"Getting repo {repo_full_name}")
    file_comment = pr_path and pr_line_position
    try:
        g = get_github_client(installation_id)
        repo = g.get_repo(repo_full_name)
        pr = repo.get_pull(pr_number)
        try:
            if not comment.lower().startswith('sweep:'):
                return
            if not comment_id:
                pass
            else:
                try:
                    item_to_react_to = pr.get_issue_comment(comment_id)
                    item_to_react_to.create_reaction("eyes")
                except Exception as e:
                    pass
                try:
                    item_to_react_to = pr.get_review_comment(comment_id)
                    item_to_react_to.create_reaction("eyes")
                except Exception as e:
                    pass
        except Exception as e:
            logger.error(f"Failed to fetch comments: {str(e)}")
        # Check if the PR is closed
        if pr.state == "closed":
            return {"success": True, "message": "PR is closed. No event fired."}
        branch_name = pr.head.ref
        pr_title = pr.title
        pr_body = pr.body or ""
        diffs = get_pr_diffs(repo, pr)
        pr_line = None
        pr_file_path = None
        # This means it's a comment on a file
        if file_comment:
            pr_file = repo.get_contents(pr_path, ref=branch_name).decoded_content.decode("utf-8")
            pr_lines = pr_file.splitlines()
            pr_line = pr_lines[min(len(pr_lines), pr_line_position) - 1]
            pr_file_path = pr_path.strip()
        # This means it's a comment on the PR
        else:
            if not comment.strip().lower().startswith("sweep"):
                logger.info("No event fired because it doesn't start with Sweep.")
                return {"success": True, "message": "No event fired."}

        def fetch_file_contents_with_retry():
            retries = 3
            error = None
            for i in range(retries):
                try:
                    logger.info(f"Fetching relevant files for the {i}th time...")
                    return search_snippets(
                        repo,
                        f"{comment}\n{pr_title}" + (f"\n{pr_line}" if pr_line else ""),
                        num_files=30,
                        branch=branch_name,
                        installation_id=installation_id,
                    )
                except Exception as e:
                    error = e
                    continue
            posthog.capture(
                username, "fetching_failed", properties={"error": error, **metadata}
            )
            raise error

        if file_comment:
            snippets = []
            tree = ""
        else:
            try:
                logger.info("Fetching relevant files...")
                snippets, tree = fetch_file_contents_with_retry()
                assert len(snippets) > 0
            except Exception as e:
                logger.error(traceback.format_exc())
                raise e
        chat_logger = ChatLogger({
            'repo_name': repo_name,
            'title': '(Comment) ' + pr_title,
            "issue_url": pr.html_url,
            "pr_file_path": pr_file_path,  # may be None
            "pr_line": pr_line,  # may be None
            "repo_full_name": repo_full_name,
            "repo_description": repo_description,
            "comment": comment,
            "pr_path": pr_path,
            "pr_line_position": pr_line_position,
            "username": username,
            "installation_id": installation_id,
            "pr_number": pr_number,
            "type": "comment",
        })
        snippets = post_process_snippets(snippets, max_num_of_snippets=0 if file_comment else 2)

        logger.info("Getting response from ChatGPT...")
        human_message = HumanMessageCommentPrompt(
            comment=comment,
            repo_name=repo_name,
            repo_description=repo_description if repo_description else "",
            diffs=diffs,
            issue_url=pr.html_url,
            username=username,
            title=pr_title,
            tree=tree,
            summary=pr_body,
            snippets=snippets,
            pr_file_path=pr_file_path,  # may be None
            pr_line=pr_line,  # may be None
        )
        logger.info(f"Human prompt{human_message.construct_prompt()}")

        sweep_bot = SweepBot.from_system_message_content(
            # human_message=human_message, model="claude-v1.3-100k", repo=repo
            human_message=human_message, repo=repo, chat_logger=chat_logger, model="gpt-4-32k-0613"
        )
    except Exception as e:
        logger.error(traceback.format_exc())
        posthog.capture(username, "failed", properties={
            "error": str(e),
            "reason": "Failed to get files",
            **metadata
        })
        raise e

    try:
        logger.info("Fetching files to modify/create...")
        file_change_requests, create_thoughts, modify_thoughts = sweep_bot.get_files_to_change(retries=3)
        file_change_requests = sweep_bot.validate_file_change_requests(file_change_requests, branch=branch_name)
        logger.info("Making Code Changes...")
        sweep_bot.change_files_in_github(file_change_requests, branch_name)

        logger.info("Done!")
    except NoFilesException:
        posthog.capture(username, "failed", properties={
            "error": "No files to change",
            "reason": "No files to change",
            **metadata
        })
        return {"success": True, "message": "No files to change."}
    except Exception as e:
        logger.error(traceback.format_exc())
        posthog.capture(username, "failed", properties={
            "error": str(e),
            "reason": "Failed to make changes",
            **metadata
        })
        raise e

    posthog.capture(username, "success", properties={**metadata})
    logger.info("on_comment success")
    return {"success": True}


def rollback_file(repo_full_name, pr_path, installation_id, pr_number):
    g = get_github_client(installation_id)
    repo = g.get_repo(repo_full_name)
    pr = repo.get_pull(pr_number)
    branch_name = pr.head.ref

    # Get the file's content from the previous commit
    commits = repo.get_commits(sha=branch_name)
    if commits.totalCount < 2:
        current_file = repo.get_contents(pr_path, ref=commits[0].sha)
        current_file_sha = current_file.sha
        previous_content = repo.get_contents(pr_path, ref=repo.default_branch)
        previous_file_content = previous_content.decoded_content.decode("utf-8")
        repo.update_file(pr_path, "Revert file to previous commit", previous_file_content, current_file_sha,
                         branch=branch_name)
        return
    previous_commit = commits[1]

    # Get current file SHA
    current_file = repo.get_contents(pr_path, ref=commits[0].sha)
    current_file_sha = current_file.sha

    # Check if the file exists in the previous commit
    try:
        previous_content = repo.get_contents(pr_path, ref=previous_commit.sha)
        previous_file_content = previous_content.decoded_content.decode("utf-8")
        # Create a new commit with the previous file content
        repo.update_file(pr_path, "Revert file to previous commit", previous_file_content, current_file_sha,
                         branch=branch_name)
    except Exception as e:
        logger.error(traceback.format_exc())
        if e.status == 404:
            logger.warning(f"File {pr_path} was not found in previous commit {previous_commit.sha}")
        else:
            raise e



import json
import os
import shutil
import tempfile
import re
import webbrowser

import gradio as gr
from git import Repo
from github import Github
from loguru import logger

from sweepai.app.api_client import APIClient, create_pr_function, create_pr_function_call
from sweepai.app.config import State, SweepChatConfig
from sweepai.core.entities import Snippet
from sweepai.utils.config.client import SweepConfig

def main():
    config = SweepChatConfig.load()

    api_client = APIClient(config=config)

    pr_summary_template = '''⏳ I'm creating the following PR...

    **{title}**
    {summary}

    Here is my plan:
    {plan}

    Reply to propose changes to the plan.'''

    print("Getting list of repos...")
    github_client = Github(config.github_pat)
    repos = list(github_client.get_user().get_repos())
    print("Done.")

    css = '''
    footer {
        visibility: hidden;
    }
    pre, code {
        white-space: pre-wrap !important;
        word-break: break-all !important;
    }
    #snippets {
        height: 600px;
        overflow-y: scroll;
    }
    #message_box > label > span {
        display: none;
    }
    '''


    def get_files_recursively(root_path, path=''):
        files = []
        path_to_contents = {}

        if path == '.git':
            return files, path_to_contents

        current_dir = os.path.join(root_path, path)
        entries = os.listdir(current_dir)

        for entry in entries:
            entry_path = os.path.join(current_dir, entry)

            if os.path.isfile(entry_path):
                try:
                    with open(entry_path, 'r', encoding="utf-8", errors="ignore") as file:
                        contents = file.read()
                    path_to_contents[entry_path[len(root_path) + 1:]] = contents
                    files.append(entry_path[len(root_path) + 1:])
                except UnicodeDecodeError as e:
                    logger.warning(f"Received warning {e}, skipping...")
                    continue
            elif os.path.isdir(entry_path):
                subfiles, subpath_to_contents = get_files_recursively(root_path, os.path.join(path, entry))
                files.extend(subfiles)
                path_to_contents.update(subpath_to_contents)

        return files, path_to_contents


    def get_installation_id(repo_full_name):
        config.repo_full_name = repo_full_name
        api_client.config = config
        installation_id = api_client.get_installation_id()
        return installation_id


    path_to_contents = {}


    def get_files(repo_full_name):
        global path_to_contents
        global repo
        if repo_full_name is None:
            all_files = []
        else:
            # Make sure repo is added to Sweep before checking all recursive files
            try:
                installation_id = get_installation_id(repo_full_name)
                assert installation_id
            except:
                return []
            repo = github_client.get_repo(repo_full_name)
            branch_name = SweepConfig.get_branch(repo)
            repo_url = f"https://x-access-token:{config.github_pat}@github.com/{repo_full_name}.git"
            try:
                repo_dir = os.path.join(tempfile.gettempdir(), repo_full_name)
                if os.path.exists(repo_dir):
                    git_repo = Repo(repo_dir)
                else:
                    git_repo = Repo.clone_from(repo_url, repo_dir)
                git_repo.git.checkout(branch_name)
                git_repo.remotes.origin.pull()
            except Exception as e:
                logger.warning(f"Git pull failed with error {e}, deleting cache and recloning...")
                shutil.rmtree(repo_dir)
                git_repo = Repo.clone_from(repo_url, repo_dir)
                git_repo.git.checkout(branch_name)
                git_repo.remotes.origin.pull()
            all_files, path_to_contents = get_files_recursively(repo_dir)
        return all_files


    def get_files_update(*args):
        global repo
        if len(args) > 0:
            repo = args[0]
        else:
            repo = config.repo_full_name
        return gr.Dropdown.update(choices=get_files(repo))


    def parse_response(raw_response: str) -> tuple[str, list[tuple[str, str]]]:
        if "Plan:" not in raw_response:
            response, raw_plan = raw_response, ""
        else:
            response, raw_plan = raw_response.split("Plan:", 1)
        if response.startswith("Response:"):
            response = response[len("Response:"):]
        plan = [(line[:line.find(":")].strip(), line[line.find(":") + 1:].strip()) for line in raw_plan.split("\n*") if
                line]
        return response, plan

    try:
        user_info = api_client.get_user_info()
    except Exception as e:
        logger.warning(e)
        user_info = {"is_paying_user": False, "remaining_tickets": 0}
    global_state = config.state

    with gr.Blocks(theme=gr.themes.Soft(), title="Sweep Chat", css=css) as demo:
        print("Launching gradio!")
        with gr.Row():
            with gr.Column(scale=2):
                repo_full_name = gr.Dropdown(choices=[repo.full_name for repo in repos], label="Repo full name",
                                            value=lambda: config.repo_full_name or "")
            print("Indexing files...")
            with gr.Column(scale=4):
                file_names = gr.Dropdown(choices=get_files(config.repo_full_name), multiselect=True, label="Files",
                                        value=lambda: global_state.file_paths)
            print("Indexed files!")
            repo_full_name.change(fn=get_files_update, inputs=repo_full_name, outputs=file_names)
            with gr.Column(scale=1):
                restart_button = gr.Button("Restart")

        with gr.Row():
            with gr.Column(scale=2):
                chatbot = gr.Chatbot(height=600, value=lambda: global_state.chat_history)
            with gr.Column():
                with gr.Row():
                    snippets_text = gr.Markdown(value=lambda: global_state.snippets_text, elem_id="snippets")

        with gr.Row():
            plan = gr.List(
                value=[[filename + ": " + instructions] for filename, instructions in global_state.plan] or [[""]],
                headers=["Proposed Plan"],
                interactive=True,
                col_count=(1, "static"),
                wrap=True,
                visible=global_state.plan_toggle,
            )

        with gr.Row():
            with gr.Column(scale=8):
                msg = gr.Textbox(placeholder="Send a message to Sweep", label=None, elem_id="message_box")
            create_pr_button_column = gr.Column(scale=0.5, visible=bool(global_state.plan))
            with create_pr_button_column:
                create_pr_button = gr.Button(value="Create PR", interactive=bool(global_state.plan))
        
        with gr.Row():
            with gr.Column():
                plan_toggle = gr.Checkbox(label="Propose plan (Experimental)", value=lambda: global_state.plan_toggle, interactive=True, elem_id="plan_toggle")
            with gr.Column(scale=5):
                tier_message = "⚡ Free Trial" if not user_info["is_paying_user"] else "💎 Pro Tier"
                remaining_tickets = user_info["remaining_tickets"]
                user_status = gr.Markdown(value=f"Logged in as {config.github_username} ({tier_message}), {remaining_tickets} tickets remaining.")
        
        def on_plan_toggle_click(plan_toggle):
            global global_state
            global_state.plan_toggle = plan_toggle
            config.state = global_state
            config.save()
            return gr.update(visible=plan_toggle), gr.update(visible=plan_toggle)
        
        plan_toggle.change(on_plan_toggle_click, [plan_toggle], [create_pr_button_column, plan])

        def clear_inputs():
            global global_state
            global_state = State()
            config.state = global_state
            config.save()
            return [], [], [[""]]


        restart_button.click(clear_inputs, None, [file_names, chatbot, plan])

        file_names.change(fn=get_files_update, inputs=repo_full_name, outputs=file_names)

        global searched
        global selected_snippets
        global file_to_str

        searched = False
        selected_snippets = []
        file_to_str = {}


        def repo_name_change(repo_full_name):
            global installation_id
            try:
                try:
                    installation_id = get_installation_id(repo_full_name)
                except Exception as e:
                    raise gr.Error(str(e))
                assert installation_id
                config.installation_id = installation_id
                api_client.config = config
                config.save()
                return ""
            except Exception as e:
                config.repo_full_name = None
                config.installation_id = None
                config.save()
                api_client.config = config
                raise e


        def build_string():
            global selected_snippets
            global file_to_str
            for snippet in selected_snippets:
                file_name = snippet.file_path
                if file_name not in file_to_str:
                    add_file_to_dict(file_name)
            snippets_text = "### Relevant snippets:\n" + "\n\n".join(
                [file_to_str[snippet.file_path] for snippet in selected_snippets])
            return snippets_text


        repo_full_name.change(fn=repo_name_change, inputs=[repo_full_name], outputs=[msg])


        def add_file_to_dict(file_name):
            global file_to_str
            global path_to_contents
            global repo
            if file_name in path_to_contents:
                file_contents = path_to_contents[file_name]
            else:
                file_contents = repo.get_contents(file_name, ref=SweepConfig.get_branch(repo)).decoded_content.decode(
                    'utf-8')
            file_contents_split = file_contents.split("\n")
            length = len(file_contents_split)
            backtick, escaped_backtick = "`", "\\`"
            preview = "\n".join(file_contents_split[:3]).replace(backtick, escaped_backtick)
            file_to_str[file_name] = f'{file_name}:0:{length}\n```\n{preview}\n...\n```'


        def file_names_change(file_names):
            global selected_snippets
            global file_to_str
            global path_to_contents
            selected_snippets = []
            for file_name in file_names:
                if file_name in path_to_contents:
                    selected_snippets.append(
                        Snippet(content=path_to_contents[file_name], start=0, end=path_to_contents[file_name].count('\n'),
                                file_path=file_name))
            return file_names, build_string()


        file_names.change(fn=file_names_change, inputs=[file_names], outputs=[file_names, snippets_text])

        def handle_message_submit(repo_full_name: str, user_message: str, history: list[tuple[str | None, str | None]]):
            if not repo_full_name:
                raise gr.Error("Set the repository name first")
            return gr.update(value="", interactive=False), history + [[user_message, None]], gr.Button.update(
                interactive=False)

        def _handle_message_stream(
            chat_history: list[tuple[str | None, str | None]], 
            snippets_text, 
            file_names, 
            plan,
            plan_toggle
        ):
            global selected_snippets
            global searched
            if plan is None or plan == [[]] or plan == [[""]] or plan == [["", ""]]:
                plan = [["", ""]]
            message = chat_history[-1][0]
            yield chat_history, snippets_text, file_names, plan
            if not selected_snippets:
                searched = True
                # Searching for relevant snippets
                chat_history[-1][1] = "Searching for relevant snippets..."
                snippets_text = build_string()
                yield chat_history, snippets_text, file_names, plan
                logger.info("Fetching relevant snippets...")
                try:
                    selected_snippets += api_client.search(chat_history[-1][0], 3)
                except Exception as e:
                    raise gr.Error(str(e))
                snippets_text = build_string()
                file_names = [snippet.file_path for snippet in selected_snippets]
                yield chat_history, snippets_text, file_names, plan
                logger.info("Fetched relevant snippets.")
                chat_history[-1][1] = "Found relevant snippets."
                # Update using chat_history
                snippets_text = build_string()
                yield chat_history, snippets_text, file_names, plan

            # Generate response
            logger.info("...")
            chat_history.append([None, "..."])
            yield chat_history, snippets_text, file_names, plan
            chat_history[-1][1] = ""
            logger.info("Starting to generate response...")
            if len(chat_history) > 1 and "create pr" in message.lower():
                try:
                    stream = api_client.stream_chat(
                        chat_history,
                        selected_snippets,
                        do_add_plan=plan_toggle,
                        functions=[create_pr_function],
                        function_call=create_pr_function_call,
                    )
                except Exception as e:
                    raise gr.Error(str(e))
            else:
                stream = api_client.stream_chat(chat_history, selected_snippets, do_add_plan=plan_toggle)
            function_name = ""
            raw_arguments = ""
            raw_response = ""
            parsed_response = ""
            try:
                for chunk in stream:
                    if chunk.get("content"):
                        token = chunk["content"]
                        raw_response += token
                        if plan_toggle:
                            parsed_response, plan = parse_response(raw_response)
                            chat_history[-1][1] = parsed_response
                            yield chat_history, snippets_text, file_names, plan
                        else:
                            chat_history[-1][1] = raw_response
                            yield chat_history, snippets_text, file_names, plan
                    if chunk.get("function_call"):
                        function_call = chunk["function_call"]
                        function_name = function_name or function_call.get("name")
                        raw_arguments += function_call.get("arguments")
                        chat_history[-1][1] = f"Calling function: `{function_name}`\n```json\n{raw_arguments}\n```"
                        yield chat_history, snippets_text, file_names, plan
            except Exception as e:
                logger.error(e)
                raise gr.Error(str(e))
            if function_name:
                arguments = json.loads(raw_arguments)
                if function_name == "create_pr":
                    assert "title" in arguments
                    assert "summary" in arguments
                    assert "plan" in arguments
                    if "branch" not in arguments:
                        arguments["branch"] = arguments["title"].lower().replace(" ", "_").replace("-", "_")[:50]
                    chat_history[-1][1] = pr_summary_template.format(
                        title=arguments["title"],
                        summary=arguments["summary"],
                        plan="\n".join([f"* `{item['file_path']}`: {item['instructions']}" for item in arguments["plan"]])
                    )
                    yield chat_history, snippets_text, file_names, plan
                    plan = [(item["file_path"], item["instructions"]) for item in arguments["plan"]]
                    yield chat_history, snippets_text, file_names, plan
                else:
                    raise NotImplementedError


        def handle_message_stream(
            chat_history: list[tuple[str | None, str | None]], 
            snippets_text, 
            file_paths, 
            plan, 
            plan_toggle
        ):
            global global_state
            for chat_history, snippets_text, file_paths, plan in _handle_message_stream(chat_history, snippets_text,
                                                                                        file_paths, plan, plan_toggle):
                if plan is None or plan == [[]] or plan == [[""]] or plan == [["", ""]]:
                    plan = [["", ""]]
                if plan and isinstance(plan[0], list):
                    plan = [item.split(":") for item in plan if ":" in item]
                global_state = State(
                    chat_history=chat_history,
                    snippets_text=snippets_text,
                    file_paths=file_paths,
                    plan=plan,
                )
                config.state = global_state
                config.save()
                if plan == [] or plan == [["", ""]]:
                    yield chat_history, snippets_text, file_paths, [[""]]
                else:
                    yield chat_history, snippets_text, file_paths, [(file_path + ": " + instructions,) for
                                                                file_path, instructions in plan]


        response = msg \
            .submit(handle_message_submit, [repo_full_name, msg, chatbot], [msg, chatbot, create_pr_button], queue=False) \
            .then(handle_message_stream, [chatbot, snippets_text, file_names, plan, plan_toggle],
                [chatbot, snippets_text, file_names, plan]) \
            .then(lambda: [gr.update(interactive=True), gr.update(interactive=True)], None, [msg, create_pr_button], queue=False)


        def validate_branch_name(branch_name):
            # Replace any characters that are not alphanumeric or '-' or '_' with '_'
            valid_branch_name = re.sub('[^0-9a-zA-Z_-]', '_', branch_name)
            return valid_branch_name

        def on_create_pr_button_click(chat_history: list[tuple[str | None, str | None]], plan: list[tuple[str]]):
            title = chat_history[0][0][:240]
            content = chat_history[-1][1]
            valid_branch_name = validate_branch_name(title.lower().replace(" ", "_").replace("-", "_")[:50])
            chat_history.append((None, "⌛ Creating PR..."))
            yield chat_history, gr.Button.update(interactive=False)
            try:
                pull_request = api_client.create_pr(
                    file_change_requests=[(item[:item.find(":")], item[item.find(":") + 1:]) for item, *_ in plan],
                    pull_request={
                        "title": title,
                        "content": content,
                        "branch_name": valid_branch_name
                    },
                    messages=chat_history,
                )
                chat_history.append((None, f"✅ PR created at {pull_request['html_url']}"))
                webbrowser.open_new_tab(pull_request["html_url"])
                yield chat_history, gr.Button.update(interactive=True)
            except Exception as e:
                yield chat_history, gr.Button.update(interactive=True)
                raise gr.Error(str(e))


        create_pr_button.click(on_create_pr_button_click, [chatbot, plan], [chatbot, create_pr_button])

    demo.queue()
    demo.launch(inbrowser=True)

if __name__ == "__main__":
    main()
